import unittest
from ipaddress import AddressValueError

from pyrad2 import tools


class EncodingTests(unittest.TestCase):
    def testStringEncoding(self):
        self.assertRaises(ValueError, tools.EncodeString, "x" * 254)
        self.assertEqual(tools.EncodeString("1234567890"), b"1234567890")

    def testInvalidStringEncodingRaisesTypeError(self):
        self.assertRaises(TypeError, tools.EncodeString, 1)

    def testAddressEncoding(self):
        self.assertRaises(AddressValueError, tools.EncodeAddress, "TEST123")
        self.assertEqual(tools.EncodeAddress("192.168.0.255"), b"\xc0\xa8\x00\xff")

    def testInvalidAddressEncodingRaisesTypeError(self):
        self.assertRaises(TypeError, tools.EncodeAddress, 1)

    def testIntegerEncoding(self):
        self.assertEqual(tools.EncodeInteger(0x01020304), b"\x01\x02\x03\x04")

    def testInteger64Encoding(self):
        self.assertEqual(tools.EncodeInteger64(0xFFFFFFFFFFFFFFFF), b"\xff" * 8)

    def testUnsignedIntegerEncoding(self):
        self.assertEqual(tools.EncodeInteger(0xFFFFFFFF), b"\xff\xff\xff\xff")

    def testInvalidIntegerEncodingRaisesTypeError(self):
        self.assertRaises(TypeError, tools.EncodeInteger, "ONE")

    def testDateEncoding(self):
        self.assertEqual(tools.EncodeDate(0x01020304), b"\x01\x02\x03\x04")

    def testInvalidDataEncodingRaisesTypeError(self):
        self.assertRaises(TypeError, tools.EncodeDate, "1")

    def testEncodeAscendBinary(self):
        self.assertEqual(
            tools.EncodeAscendBinary(
                "family=ipv4 action=discard direction=in dst=10.10.255.254/32"
            ),
            b"\x01\x00\x01\x00\x00\x00\x00\x00\n\n\xff\xfe\x00 \x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00",
        )

    def testStringDecoding(self):
        self.assertEqual(tools.DecodeString(b"1234567890"), "1234567890")

    def testAddressDecoding(self):
        self.assertEqual(tools.DecodeAddress(b"\xc0\xa8\x00\xff"), "192.168.0.255")

    def testIntegerDecoding(self):
        self.assertEqual(tools.DecodeInteger(b"\x01\x02\x03\x04"), 0x01020304)

    def testInteger64Decoding(self):
        self.assertEqual(tools.DecodeInteger64(b"\xff" * 8), 0xFFFFFFFFFFFFFFFF)

    def testDateDecoding(self):
        self.assertEqual(tools.DecodeDate(b"\x01\x02\x03\x04"), 0x01020304)

    def testOctetsEncoding(self):
        self.assertEqual(tools.EncodeOctets("0x01020304"), b"\x01\x02\x03\x04")
        self.assertEqual(tools.EncodeOctets(b"0x01020304"), b"\x01\x02\x03\x04")
        self.assertEqual(tools.EncodeOctets("16909060"), b"\x01\x02\x03\x04")
        # encodes to 253 bytes
        self.assertEqual(
            tools.EncodeOctets(
                "0x0102030405060708090A0B0C0D0E0F100102030405060708090A0B0C0D0E0F100102030405060708090A0B0C0D0E0F100102030405060708090A0B0C0D0E0F100102030405060708090A0B0C0D0E0F100102030405060708090A0B0C0D0E0F100102030405060708090A0B0C0D0E0F100102030405060708090A0B0C0D0E0F100102030405060708090A0B0C0D0E0F100102030405060708090A0B0C0D0E0F100102030405060708090A0B0C0D0E0F100102030405060708090A0B0C0D0E0F100102030405060708090A0B0C0D0E0F100102030405060708090A0B0C0D0E0F100102030405060708090A0B0C0D0E0F100102030405060708090A0B0C0D"
            ),
            b"\x01\x02\x03\x04\x05\x06\x07\x08\t\n\x0b\x0c\r\x0e\x0f\x10\x01\x02\x03\x04\x05\x06\x07\x08\t\n\x0b\x0c\r\x0e\x0f\x10\x01\x02\x03\x04\x05\x06\x07\x08\t\n\x0b\x0c\r\x0e\x0f\x10\x01\x02\x03\x04\x05\x06\x07\x08\t\n\x0b\x0c\r\x0e\x0f\x10\x01\x02\x03\x04\x05\x06\x07\x08\t\n\x0b\x0c\r\x0e\x0f\x10\x01\x02\x03\x04\x05\x06\x07\x08\t\n\x0b\x0c\r\x0e\x0f\x10\x01\x02\x03\x04\x05\x06\x07\x08\t\n\x0b\x0c\r\x0e\x0f\x10\x01\x02\x03\x04\x05\x06\x07\x08\t\n\x0b\x0c\r\x0e\x0f\x10\x01\x02\x03\x04\x05\x06\x07\x08\t\n\x0b\x0c\r\x0e\x0f\x10\x01\x02\x03\x04\x05\x06\x07\x08\t\n\x0b\x0c\r\x0e\x0f\x10\x01\x02\x03\x04\x05\x06\x07\x08\t\n\x0b\x0c\r\x0e\x0f\x10\x01\x02\x03\x04\x05\x06\x07\x08\t\n\x0b\x0c\r\x0e\x0f\x10\x01\x02\x03\x04\x05\x06\x07\x08\t\n\x0b\x0c\r\x0e\x0f\x10\x01\x02\x03\x04\x05\x06\x07\x08\t\n\x0b\x0c\r\x0e\x0f\x10\x01\x02\x03\x04\x05\x06\x07\x08\t\n\x0b\x0c\r\x0e\x0f\x10\x01\x02\x03\x04\x05\x06\x07\x08\t\n\x0b\x0c\r",
        )
        self.assertRaisesRegex(
            ValueError,
            "Can only encode strings of <= 253 characters",
            tools.EncodeOctets,
            "0x0102030405060708090A0B0C0D0E0F100102030405060708090A0B0C0D0E0F100102030405060708090A0B0C0D0E0F100102030405060708090A0B0C0D0E0F100102030405060708090A0B0C0D0E0F100102030405060708090A0B0C0D0E0F100102030405060708090A0B0C0D0E0F100102030405060708090A0B0C0D0E0F100102030405060708090A0B0C0D0E0F100102030405060708090A0B0C0D0E0F100102030405060708090A0B0C0D0E0F100102030405060708090A0B0C0D0E0F100102030405060708090A0B0C0D0E0F100102030405060708090A0B0C0D0E0F100102030405060708090A0B0C0D0E0F100102030405060708090A0B0C0D0E",
        )

    def testUnknownTypeEncoding(self):
        self.assertRaises(ValueError, tools.EncodeAttr, "unknown", None)

    def testUnknownTypeDecoding(self):
        self.assertRaises(ValueError, tools.DecodeAttr, "unknown", None)

    def testEncodeFunction(self):
        self.assertEqual(tools.EncodeAttr("string", "string"), b"string")
        self.assertEqual(tools.EncodeAttr("octets", b"string"), b"string")
        self.assertEqual(
            tools.EncodeAttr("ipaddr", "192.168.0.255"), b"\xc0\xa8\x00\xff"
        )
        self.assertEqual(tools.EncodeAttr("integer", 0x01020304), b"\x01\x02\x03\x04")
        self.assertEqual(tools.EncodeAttr("date", 0x01020304), b"\x01\x02\x03\x04")
        self.assertEqual(tools.EncodeAttr("integer64", 0xFFFFFFFFFFFFFFFF), b"\xff" * 8)

    def testDecodeFunction(self):
        self.assertEqual(tools.DecodeAttr("string", b"string"), "string")
        self.assertEqual(tools.EncodeAttr("octets", b"string"), b"string")
        self.assertEqual(
            tools.DecodeAttr("ipaddr", b"\xc0\xa8\x00\xff"), "192.168.0.255"
        )
        self.assertEqual(tools.DecodeAttr("integer", b"\x01\x02\x03\x04"), 0x01020304)
        self.assertEqual(tools.DecodeAttr("integer64", b"\xff" * 8), 0xFFFFFFFFFFFFFFFF)
        self.assertEqual(tools.DecodeAttr("date", b"\x01\x02\x03\x04"), 0x01020304)



import hmac
import os
import struct
import unittest
import hashlib

from .base import TEST_ROOT_PATH

from collections import OrderedDict
from pyrad2 import packet
from pyrad2.client import Client
from pyrad2.dictionary import Dictionary
from pyrad2.constants import PacketType


class UtilityTests(unittest.TestCase):
    def testGenerateID(self):
        id = packet.CreateID()
        self.assertTrue(isinstance(id, int))
        newid = packet.CreateID()
        self.assertNotEqual(id, newid)


class PacketConstructionTests(unittest.TestCase):
    klass = packet.Packet

    def setUp(self):
        self.path = os.path.join(TEST_ROOT_PATH, "data")
        self.dict = Dictionary(os.path.join(self.path, "simple"))

    def testBasicConstructor(self):
        pkt = self.klass()
        self.assertTrue(isinstance(pkt.code, int))
        self.assertTrue(isinstance(pkt.id, int))
        self.assertTrue(isinstance(pkt.secret, bytes))

    def testNamedConstructor(self):
        pkt = self.klass(
            code=26,
            id=38,
            secret=b"secret",
            authenticator=b"authenticator",
            dict="fakedict",
        )
        self.assertEqual(pkt.code, 26)
        self.assertEqual(pkt.id, 38)
        self.assertEqual(pkt.secret, b"secret")
        self.assertEqual(pkt.authenticator, b"authenticator")
        self.assertEqual(pkt.dict, "fakedict")

    def testConstructWithDictionary(self):
        pkt = self.klass(dict=self.dict)
        self.assertTrue(pkt.dict is self.dict)

    def testConstructorIgnoredParameters(self):
        marker = []
        pkt = self.klass(fd=marker)
        self.assertFalse(getattr(pkt, "fd", None) is marker)

    def testSecretMustBeBytestring(self):
        self.assertRaises(TypeError, self.klass, secret="secret")

    def testConstructorWithAttributes(self):
        pkt = self.klass(**{"Test-String": "this works", "dict": self.dict})
        self.assertEqual(pkt["Test-String"], ["this works"])

    def testConstructorWithTlvAttribute(self):
        pkt = self.klass(
            **{"Test-Tlv-Str": "this works", "Test-Tlv-Int": 10, "dict": self.dict}
        )
        self.assertEqual(
            pkt["Test-Tlv"], {"Test-Tlv-Str": ["this works"], "Test-Tlv-Int": [10]}
        )


class PacketTests(unittest.TestCase):
    def setUp(self):
        self.path = os.path.join(TEST_ROOT_PATH, "data")
        self.dict = Dictionary(os.path.join(self.path, "full"))
        self.packet = packet.Packet(
            id=0, secret=b"secret", authenticator=b"01234567890ABCDEF", dict=self.dict
        )

    def _create_reply_with_duplicate_attributes(self, request):
        """
        Creates a reply to the given request with multiple instances of the
        same attribute that also do not appear sequentially in the list. Used
        to ensure that methods providing authenticator and
        Message-Authenticator verification can handle the case where multiple
        instances of an given attribute do not appear sequentially in the
        attributes list.
        """
        # Manually build the packet since using packet.Packet will always group
        # attributes of the same type together
        attributes = self._get_attribute_bytes("Test-String", "test")
        attributes += self._get_attribute_bytes("Test-Integer", 1)
        attributes += self._get_attribute_bytes("Test-String", "test")
        attributes += self._get_attribute_bytes("Message-Authenticator", 16 * b"\00")

        header = struct.pack(
            "!BBH", PacketType.AccessAccept, request.id, (20 + len(attributes))
        )

        # Calculate the Message-Authenticator and update the attribute
        hmac_constructor = hmac.new(request.secret, None, hashlib.md5)
        hmac_constructor.update(header + request.authenticator + attributes)
        updated_message_authenticator = hmac_constructor.digest()
        attributes = attributes.replace(b"\x00" * 16, updated_message_authenticator)

        # Calculate the response authenticator
        authenticator = hashlib.md5(
            header + request.authenticator + attributes + request.secret
        ).digest()

        reply_bytes = header + authenticator + attributes
        return packet.AuthPacket(packet=reply_bytes, dict=self.dict)

    def _get_attribute_bytes(self, attr_name, value):
        attr = self.dict.attributes[attr_name]
        attr_key = attr.code
        attr_value = packet.tools.EncodeAttr(attr.type, value)
        attr_len = len(attr_value) + 2
        return struct.pack("!BB", attr_key, attr_len) + attr_value

    def testCreateReply(self):
        reply = self.packet.CreateReply(**{"Test-Integer": 10})
        self.assertEqual(reply.id, self.packet.id)
        self.assertEqual(reply.secret, self.packet.secret)
        self.assertEqual(reply.authenticator, self.packet.authenticator)
        self.assertEqual(reply["Test-Integer"], [10])

    def testAttributeAccess(self):
        self.packet["Test-Integer"] = 10
        self.assertEqual(self.packet["Test-Integer"], [10])
        self.assertEqual(self.packet[3], [b"\x00\x00\x00\x0a"])

        self.packet["Test-String"] = "dummy"
        self.assertEqual(self.packet["Test-String"], ["dummy"])
        self.assertEqual(self.packet[1], [b"dummy"])

    def testAttributeValueAccess(self):
        self.packet["Test-Integer"] = "Three"
        self.assertEqual(self.packet["Test-Integer"], ["Three"])
        self.assertEqual(self.packet[3], [b"\x00\x00\x00\x03"])

    def testVendorAttributeAccess(self):
        self.packet["Simplon-Number"] = 10
        self.assertEqual(self.packet["Simplon-Number"], [10])
        self.assertEqual(self.packet[(16, 1)], [b"\x00\x00\x00\x0a"])

        self.packet["Simplon-Number"] = "Four"
        self.assertEqual(self.packet["Simplon-Number"], ["Four"])
        self.assertEqual(self.packet[(16, 1)], [b"\x00\x00\x00\x04"])

    def testRawAttributeAccess(self):
        marker = [b""]
        self.packet[1] = marker
        self.assertTrue(self.packet[1] is marker)
        self.packet[(16, 1)] = marker
        self.assertTrue(self.packet[(16, 1)] is marker)

    def testEncryptedAttributes(self):
        self.packet["Test-Encrypted-String"] = "dummy"
        self.assertEqual(self.packet["Test-Encrypted-String"], ["dummy"])

        self.packet["Test-Encrypted-Integer"] = 10
        self.assertEqual(self.packet["Test-Encrypted-Integer"], [10])

    def testHasKey(self):
        self.assertEqual(self.packet.has_key("Test-String"), False)
        self.assertEqual("Test-String" in self.packet, False)
        self.packet["Test-String"] = "dummy"
        self.assertEqual(self.packet.has_key("Test-String"), True)
        self.assertEqual(self.packet.has_key(1), True)
        self.assertEqual(1 in self.packet, True)

    def testHasKeyWithUnknownKey(self):
        self.assertEqual(self.packet.has_key("Unknown-Attribute"), False)
        self.assertEqual("Unknown-Attribute" in self.packet, False)

    def testDelItem(self):
        self.packet["Test-String"] = "dummy"
        del self.packet["Test-String"]
        self.assertEqual(self.packet.has_key("Test-String"), False)
        self.packet["Test-String"] = "dummy"
        del self.packet[1]
        self.assertEqual(self.packet.has_key("Test-String"), False)

    def testKeys(self):
        self.assertEqual(self.packet.keys(), [])
        self.packet["Test-String"] = "dummy"
        self.assertEqual(self.packet.keys(), ["Test-String"])
        self.packet["Test-Integer"] = 10
        self.assertEqual(self.packet.keys(), ["Test-String", "Test-Integer"])
        OrderedDict.__setitem__(self.packet, 12345, None)
        self.assertEqual(self.packet.keys(), ["Test-String", "Test-Integer", 12345])

    def testCreateAuthenticator(self):
        a = packet.Packet.CreateAuthenticator()
        self.assertTrue(isinstance(a, bytes))
        self.assertEqual(len(a), 16)

        b = packet.Packet.CreateAuthenticator()
        self.assertNotEqual(a, b)

    def testGenerateID(self):
        id = self.packet.CreateID()
        self.assertTrue(isinstance(id, int))
        newid = self.packet.CreateID()
        self.assertNotEqual(id, newid)

    def testReplyPacket(self):
        reply = self.packet.ReplyPacket()
        self.assertEqual(
            reply,
            (
                b"\x00\x00\x00\x14\xb0\x5e\x4b\xfb\xcc\x1c"
                b"\x8c\x8e\xc4\x72\xac\xea\x87\x45\x63\xa7"
            ),
        )

    def testVerifyReply(self):
        reply = self.packet.CreateReply()
        self.assertEqual(self.packet.VerifyReply(reply), True)

        reply.id += 1
        self.assertEqual(self.packet.VerifyReply(reply), False)
        reply.id = self.packet.id

        reply.secret = b"different"
        self.assertEqual(self.packet.VerifyReply(reply), False)
        reply.secret = self.packet.secret

        reply.authenticator = b"X" * 16
        self.assertEqual(self.packet.VerifyReply(reply), False)
        reply.authenticator = self.packet.authenticator

    def testVerifyReplyDuplicateAttributes(self):
        reply = self._create_reply_with_duplicate_attributes(self.packet)
        self.assertTrue(self.packet.VerifyReply(reply=reply, rawreply=reply.raw_packet))

    def testVerifyMessageAuthenticator(self):
        reply = self.packet.CreateReply(
            **{
                "Test-String": "test",
                "Test-Integer": 3,
            }
        )
        reply.code = PacketType.AccessAccept
        reply.add_message_authenticator()
        reply._refresh_message_authenticator()
        self.assertTrue(
            reply.verify_message_authenticator(
                secret=b"secret",
                original_authenticator=self.packet.authenticator,
                original_code=self.packet.code,
            )
        )

        self.assertFalse(
            reply.verify_message_authenticator(
                secret=b"bad_secret",
                original_authenticator=self.packet.authenticator,
                original_code=self.packet.code,
            )
        )

        self.assertFalse(
            reply.verify_message_authenticator(
                secret=b"secret",
                original_authenticator=b"bad_authenticator",
                original_code=self.packet.code,
            )
        )

    def testVerifyMessageAuthenticatorDuplicateAttributes(self):
        reply = self._create_reply_with_duplicate_attributes(self.packet)
        self.assertTrue(
            reply.verify_message_authenticator(
                secret=b"secret",
                original_authenticator=self.packet.authenticator,
                original_code=PacketType.AccessRequest,
            )
        )

    def testPktEncodeAttribute(self):
        encode = self.packet._PktEncodeAttribute

        # Encode a normal attribute
        self.assertEqual(encode(1, b"value"), b"\x01\x07value")
        # Encode a vendor attribute
        self.assertEqual(
            encode((1, 2), b"value"), b"\x1a\x0d\x00\x00\x00\x01\x02\x07value"
        )

    def testPktEncodeTlvAttribute(self):
        encode = self.packet._PktEncodeTlv

        # Encode a normal tlv attribute
        self.assertEqual(
            encode(4, {1: [b"value"], 2: [b"\x00\x00\x00\x02"]}),
            b"\x04\x0f\x01\x07value\x02\x06\x00\x00\x00\x02",
        )

        # Encode a normal tlv attribute with several sub attribute instances
        self.assertEqual(
            encode(4, {1: [b"value", b"other"], 2: [b"\x00\x00\x00\x02"]}),
            b"\x04\x16\x01\x07value\x02\x06\x00\x00\x00\x02\x01\x07other",
        )
        # Encode a vendor tlv attribute
        self.assertEqual(
            encode((16, 3), {1: [b"value"], 2: [b"\x00\x00\x00\x02"]}),
            b"\x1a\x15\x00\x00\x00\x10\x03\x0f\x01\x07value\x02\x06\x00\x00\x00\x02",
        )

    def testPktEncodeLongTlvAttribute(self):
        encode = self.packet._PktEncodeTlv

        long_str = b"a" * 245
        # Encode a long tlv attribute - check it is split between AVPs
        self.assertEqual(
            encode(4, {1: [b"value", long_str], 2: [b"\x00\x00\x00\x02"]}),
            b"\x04\x0f\x01\x07value\x02\x06\x00\x00\x00\x02\x04\xf9\x01\xf7" + long_str,
        )

        # Encode a long vendor tlv attribute
        first_avp = (
            b"\x1a\x15\x00\x00\x00\x10\x03\x0f\x01\x07value\x02\x06\x00\x00\x00\x02"
        )
        second_avp = b"\x1a\xff\x00\x00\x00\x10\x03\xf9\x01\xf7" + long_str
        self.assertEqual(
            encode((16, 3), {1: [b"value", long_str], 2: [b"\x00\x00\x00\x02"]}),
            first_avp + second_avp,
        )

    def testPktEncodeAttributes(self):
        self.packet[1] = [b"value"]
        self.assertEqual(self.packet._PktEncodeAttributes(), b"\x01\x07value")

        self.packet.clear()
        self.packet[(16, 2)] = [b"value"]
        self.assertEqual(
            self.packet._PktEncodeAttributes(), b"\x1a\x0d\x00\x00\x00\x10\x02\x07value"
        )

        self.packet.clear()
        self.packet[1] = [b"one", b"two", b"three"]
        self.assertEqual(
            self.packet._PktEncodeAttributes(), b"\x01\x05one\x01\x05two\x01\x07three"
        )

        self.packet.clear()
        self.packet[1] = [b"value"]
        self.packet[(16, 2)] = [b"value"]
        self.assertEqual(
            self.packet._PktEncodeAttributes(),
            b"\x01\x07value\x1a\x0d\x00\x00\x00\x10\x02\x07value",
        )

    def testPktDecodeVendorAttribute(self):
        decode = self.packet._PktDecodeVendorAttribute

        # Non-RFC2865 recommended form
        self.assertEqual(decode(b""), [(26, b"")])
        self.assertEqual(decode(b"12345"), [(26, b"12345")])

        # Almost RFC2865 recommended form: bad length value
        self.assertEqual(
            decode(b"\x00\x00\x00\x01\x02\x06value"),
            [(26, b"\x00\x00\x00\x01\x02\x06value")],
        )

        # Proper RFC2865 recommended form
        self.assertEqual(
            decode(b"\x00\x00\x00\x10\x02\x07value"), [((16, 2), b"value")]
        )

    def testPktDecodeTlvAttribute(self):
        decode = self.packet._PktDecodeTlvAttribute

        decode(4, b"\x01\x07value")
        self.assertEqual(self.packet[4], {1: [b"value"]})

        # add another instance of the same sub attribute
        decode(4, b"\x01\x07other")
        self.assertEqual(self.packet[4], {1: [b"value", b"other"]})

        # add a different sub attribute
        decode(4, b"\x02\x07\x00\x00\x00\x01")
        self.assertEqual(
            self.packet[4], {1: [b"value", b"other"], 2: [b"\x00\x00\x00\x01"]}
        )

    def testDecodePacketWithEmptyPacket(self):
        try:
            self.packet.DecodePacket(b"")
        except packet.PacketError as e:
            self.assertTrue("header is corrupt" in str(e))
        else:
            self.fail()

    def testDecodePacketWithInvalidLength(self):
        try:
            self.packet.DecodePacket(b"\x00\x00\x00\x001234567890123456")
        except packet.PacketError as e:
            self.assertTrue("invalid length" in str(e))
        else:
            self.fail()

    def testDecodePacketWithTooBigPacket(self):
        try:
            self.packet.DecodePacket(b"\x00\x00\x24\x00" + (0x2400 - 4) * b"X")
        except packet.PacketError as e:
            self.assertTrue("too long" in str(e))
        else:
            self.fail()

    def testDecodePacketWithPartialAttributes(self):
        try:
            self.packet.DecodePacket(b"\x01\x02\x00\x151234567890123456\x00")
        except packet.PacketError as e:
            self.assertTrue("header is corrupt" in str(e))
        else:
            self.fail()

    def testDecodePacketWithoutAttributes(self):
        self.packet.DecodePacket(b"\x01\x02\x00\x141234567890123456")
        self.assertEqual(self.packet.code, 1)
        self.assertEqual(self.packet.id, 2)
        self.assertEqual(self.packet.authenticator, b"1234567890123456")
        self.assertEqual(self.packet.keys(), [])

    def testDecodePacketWithBadAttribute(self):
        try:
            self.packet.DecodePacket(b"\x01\x02\x00\x161234567890123456\x00\x01")
        except packet.PacketError as e:
            self.assertTrue("too small" in str(e))
        else:
            self.fail()

    def testDecodePacketWithEmptyAttribute(self):
        self.packet.DecodePacket(b"\x01\x02\x00\x161234567890123456\x01\x02")
        self.assertEqual(self.packet[1], [b""])

    def testDecodePacketWithAttribute(self):
        self.packet.DecodePacket(b"\x01\x02\x00\x1b1234567890123456\x01\x07value")
        self.assertEqual(self.packet[1], [b"value"])

    def testDecodePacketWithTlvAttribute(self):
        self.packet.DecodePacket(
            b"\x01\x02\x00\x1d1234567890123456\x04\x09\x01\x07value"
        )
        self.assertEqual(self.packet[4], {1: [b"value"]})

    def testDecodePacketWithVendorTlvAttribute(self):
        self.packet.DecodePacket(
            b"\x01\x02\x00\x231234567890123456\x1a\x0f\x00\x00\x00\x10\x03\x09\x01\x07value"
        )
        self.assertEqual(self.packet[(16, 3)], {1: [b"value"]})

    def testDecodePacketWithTlvAttributeWith2SubAttributes(self):
        self.packet.DecodePacket(
            b"\x01\x02\x00\x231234567890123456\x04\x0f\x01\x07value\x02\x06\x00\x00\x00\x09"
        )
        self.assertEqual(self.packet[4], {1: [b"value"], 2: [b"\x00\x00\x00\x09"]})

    def testDecodePacketWithSplitTlvAttribute(self):
        self.packet.DecodePacket(
            b"\x01\x02\x00\x251234567890123456\x04\x09\x01\x07value\x04\x09\x02\x06\x00\x00\x00\x09"
        )
        self.assertEqual(self.packet[4], {1: [b"value"], 2: [b"\x00\x00\x00\x09"]})

    def testDecodePacketWithMultiValuedAttribute(self):
        self.packet.DecodePacket(
            b"\x01\x02\x00\x1e1234567890123456\x01\x05one\x01\x05two"
        )
        self.assertEqual(self.packet[1], [b"one", b"two"])

    def testDecodePacketWithTwoAttributes(self):
        self.packet.DecodePacket(
            b"\x01\x02\x00\x1e1234567890123456\x01\x05one\x01\x05two"
        )
        self.assertEqual(self.packet[1], [b"one", b"two"])

    def testDecodePacketWithVendorAttribute(self):
        self.packet.DecodePacket(b"\x01\x02\x00\x1b1234567890123456\x1a\x07value")
        self.assertEqual(self.packet[26], [b"value"])

    def testEncodeKeyValues(self):
        self.assertEqual(self.packet._EncodeKeyValues(1, "1234"), (1, "1234"))

    def testEncodeKey(self):
        self.assertEqual(self.packet._EncodeKey(1), 1)

    def testAddAttribute(self):
        self.packet.AddAttribute("Test-String", "1")
        self.assertEqual(self.packet["Test-String"], ["1"])
        self.packet.AddAttribute("Test-String", "1")
        self.assertEqual(self.packet["Test-String"], ["1", "1"])
        self.packet.AddAttribute("Test-String", ["2", "3"])
        self.assertEqual(self.packet["Test-String"], ["1", "1", "2", "3"])


class AuthPacketConstructionTests(PacketConstructionTests):
    klass = packet.AuthPacket

    def testConstructorDefaults(self):
        pkt = self.klass()
        self.assertEqual(pkt.code, PacketType.AccessRequest)


class AuthPacketTests(unittest.TestCase):
    def setUp(self):
        self.path = os.path.join(TEST_ROOT_PATH, "data")
        self.dict = Dictionary(os.path.join(self.path, "full"))
        self.packet = packet.AuthPacket(
            id=0, secret=b"secret", authenticator=b"01234567890ABCDEF", dict=self.dict
        )

    def testCreateReply(self):
        reply = self.packet.CreateReply(**{"Test-Integer": 10})
        self.assertEqual(reply.code, PacketType.AccessAccept)
        self.assertEqual(reply.id, self.packet.id)
        self.assertEqual(reply.secret, self.packet.secret)
        self.assertEqual(reply.authenticator, self.packet.authenticator)
        self.assertEqual(reply["Test-Integer"], [10])

    def testRequestPacket(self):
        self.assertEqual(
            self.packet.RequestPacket(), b"\x01\x00\x00\x1401234567890ABCDE"
        )

    def testRequestPacketCreatesAuthenticator(self):
        self.packet.authenticator = None
        self.packet.RequestPacket()
        self.assertTrue(self.packet.authenticator is not None)

    def testRequestPacketCreatesID(self):
        self.packet.id = None
        self.packet.RequestPacket()
        self.assertTrue(self.packet.id is not None)

    def testPwCryptEmptyPassword(self):
        self.assertEqual(self.packet.PwCrypt(""), b"")

    def testPwCryptPassword(self):
        self.assertEqual(
            self.packet.PwCrypt("Simplon"),
            b"\xd3U;\xb23\r\x11\xba\x07\xe3\xa8*\xa8x\x14\x01",
        )

    def testPwCryptSetsAuthenticator(self):
        self.packet.authenticator = None
        self.packet.PwCrypt("")
        self.assertTrue(self.packet.authenticator is not None)

    def testPwDecryptEmptyPassword(self):
        self.assertEqual(self.packet.PwDecrypt(b""), "")

    def testPwDecryptPassword(self):
        self.assertEqual(
            self.packet.PwDecrypt(b"\xd3U;\xb23\r\x11\xba\x07\xe3\xa8*\xa8x\x14\x01"),
            "Simplon",
        )


class AuthPacketChapTests(unittest.TestCase):
    def setUp(self):
        self.path = os.path.join(TEST_ROOT_PATH, "data")
        self.dict = Dictionary(os.path.join(self.path, "chap"))
        # self.packet = packet.Packet(id=0, secret=b'secret',
        #                             dict=self.dict)
        self.client = Client(server="localhost", secret=b"secret", dict=self.dict)

    def testVerifyChapPasswd(self):
        chap_id = b"9"
        chap_challenge = b"987654321"
        chap_password = (
            chap_id + hashlib.md5(chap_id + b"test_password" + chap_challenge).digest()
        )
        pkt = self.client.CreateAuthPacket(
            code=PacketType.AccessChallenge,
            authenticator=b"ABCDEFG",
            User_Name="test_name",
            CHAP_Challenge=chap_challenge,
            CHAP_Password=chap_password,
        )
        self.assertEqual(pkt["CHAP-Challenge"][0], chap_challenge)
        self.assertEqual(pkt["CHAP-Password"][0], chap_password)
        self.assertEqual(pkt.VerifyChapPasswd("test_password"), True)


class AcctPacketConstructionTests(PacketConstructionTests):
    klass = packet.AcctPacket

    def testConstructorDefaults(self):
        pkt = self.klass()
        self.assertEqual(pkt.code, PacketType.AccountingRequest)

    def testConstructorRawPacket(self):
        raw = (
            b"\x00\x00\x00\x14\xb0\x5e\x4b\xfb\xcc\x1c"
            b"\x8c\x8e\xc4\x72\xac\xea\x87\x45\x63\xa7"
        )
        pkt = self.klass(packet=raw)
        self.assertEqual(pkt.raw_packet, raw)


class AcctPacketTests(unittest.TestCase):
    def setUp(self):
        self.path = os.path.join(TEST_ROOT_PATH, "data")
        self.dict = Dictionary(os.path.join(self.path, "full"))
        self.packet = packet.AcctPacket(
            id=0, secret=b"secret", authenticator=b"01234567890ABCDEF", dict=self.dict
        )

    def testCreateReply(self):
        reply = self.packet.CreateReply(**{"Test-Integer": 10})
        self.assertEqual(reply.code, PacketType.AccountingResponse)
        self.assertEqual(reply.id, self.packet.id)
        self.assertEqual(reply.secret, self.packet.secret)
        self.assertEqual(reply.authenticator, self.packet.authenticator)
        self.assertEqual(reply["Test-Integer"], [10])

    def testVerifyAcctRequest(self):
        rawpacket = self.packet.RequestPacket()
        pkt = packet.AcctPacket(secret=b"secret", packet=rawpacket)
        self.assertEqual(pkt.VerifyAcctRequest(), True)

        pkt.secret = b"different"
        self.assertEqual(pkt.VerifyAcctRequest(), False)
        pkt.secret = b"secret"

        pkt.raw_packet = b"X" + pkt.raw_packet[1:]
        self.assertEqual(pkt.VerifyAcctRequest(), False)

    def testRequestPacket(self):
        self.assertEqual(
            self.packet.RequestPacket(),
            b"\x04\x00\x00\x14\x95\xdf\x90\xccbn\xfb\x15G!\x13\xea\xfa>6\x0f",
        )

    def testRequestPacketSetsId(self):
        self.packet.id = None
        self.packet.RequestPacket()
        self.assertTrue(self.packet.id is not None)



import socket
from typing import Optional

from pyrad2 import packet
from pyrad2.dictionary import Dictionary


class Host:
    """Generic RADIUS capable host."""

    def __init__(
        self,
        authport: int = 1812,
        acctport: int = 1813,
        coaport: int = 3799,
        dict: Optional[Dictionary] = None,
    ):
        """Initializes a host.

        Args:
            authport (int): port to listen on for authentication packets
            acctport (int): port to listen on for accounting packets
            coaport (int): port to listen on for CoA packets
            dict (Dictionary): RADIUS dictionary
        """
        self.dict = dict
        self.authport = authport
        self.acctport = acctport
        self.coaport = coaport

    def CreatePacket(self, **args) -> packet.Packet:
        """Create a new RADIUS packet.
        This utility function creates a new RADIUS authentication
        packet which can be used to communicate with the RADIUS server
        this client talks to. This is initializing the new packet with
        the dictionary and secret used for the client.

        Returns:
            pyrad2.packet.Packet: A new empty packet instance.
        """
        return packet.Packet(dict=self.dict, **args)

    def CreateAuthPacket(self, **args) -> packet.Packet:
        """Create a new authentication RADIUS packet.
        This utility function creates a new RADIUS authentication
        packet which can be used to communicate with the RADIUS server
        this client talks to. This is initializing the new packet with
        the dictionary and secret used for the client.

        Returns:
            pyrad2.packet.Packet: A new empty packet instance.
        """
        return packet.AuthPacket(dict=self.dict, **args)

    def CreateAcctPacket(self, **args) -> packet.Packet:
        """Create a new accounting RADIUS packet.
        This utility function creates a new accounting RADIUS packet
        which can be used to communicate with the RADIUS server this
        client talks to. This is initializing the new packet with the
        dictionary and secret used for the client.

        Returns:
            packet.Packet: A new empty packet instance.
        """
        return packet.AcctPacket(dict=self.dict, **args)

    def CreateCoAPacket(self, **args) -> packet.Packet:
        """Create a new CoA RADIUS packet.
        This utility function creates a new CoA RADIUS packet
        which can be used to communicate with the RADIUS server this
        client talks to. This is initializing the new packet with the
        dictionary and secret used for the client.

        Returns:
            packet.Packet: A new empty packet instance.
        """
        return packet.CoAPacket(dict=self.dict, **args)

    def SendPacket(self, fd: socket.socket, pkt) -> None:
        """Send a packet.

        Args:
            fd (socket.socket): Socket to send packet with
            pkt (packet.Packet): The packet instance
        """
        fd.sendto(pkt.Packet(), pkt.source)

    def SendReplyPacket(self, fd: socket.socket, pkt: packet.Packet) -> None:
        """Send a packet.

        Args:
            fd (socket.socket): Socket to send packet with
            pkt (packet.Packet): The packet instance
        """
        fd.sendto(pkt.ReplyPacket(), pkt.source)  # type: ignore



import os
if os.name == 'nt':
    import selectors
else:
    import select
import socket
from dataclasses import dataclass
from typing import Callable, Optional

from loguru import logger

from pyrad2 import host, packet
from pyrad2.dictionary import Dictionary
from pyrad2.exceptions import ServerPacketError
from pyrad2.constants import PacketType


@dataclass
class RemoteHost:
    """Remote RADIUS capable host we can talk to.

    Args:
        address (str): IP address.
        secret (bytes): RADIUS secret. If connecting to a RadSec server, the secret should be `radsec`.
        name (str): Short name (used for logging only).
        authport (int): Port used for authentication packets.
        acctport (int): Port used for accounting packets.
        coaport (int): Port used for CoA packets.
    """

    address: str
    secret: bytes
    name: str
    authport: int = 1812
    acctport: int = 1813
    coaport: int = 3799


class Server(host.Host):
    """Basic RADIUS server.
    This class implements the basics of a RADIUS server. It takes care
    of the details of receiving and decoding requests; processing of
    the requests should be done by overloading the appropriate methods
    in derived classes.

    Attributes:
        hosts (dict): Hosts who are allowed to talk to us. Dictionary of Host class instances.
        _poll (select.poll): Poll object for network sockets.
        _fdmap (dict): Map of file descriptors to network sockets.
        MaxPacketSize (int): Maximum size of a RADIUS packet. (class variable)
    """

    MaxPacketSize = 8192

    def __init__(
        self,
        addresses: Optional[list[str]] = None,
        authport: int = 1812,
        acctport: int = 1813,
        coaport: int = 3799,
        hosts: Optional[dict] = None,
        dict: Optional[Dictionary] = None,
        auth_enabled: bool = True,
        acct_enabled: bool = True,
        coa_enabled: bool = False,
    ):
        """Initializes a sync server.

        Args:
            addresses (Sequence[str]): IP addresses to listen on.
            authport (int): Port to listen on for authentication packets.
            acctport (int): Port to listen on for accounting packets.
            coaport (int): Port to listen on for CoA packets.
            hosts (dict[str, RemoteHost]): Hosts who we can talk to. A dictionary mapping IP to RemoteHost class instances.
            dict (Dictionary): RADIUS dictionary to use.
            auth_enabled (bool): Enable auth server (default: True).
            acct_enabled (bool): Enable accounting server (default: True).
            coa_enabled (bool): Enable CoA server (default: False).
        """
        super().__init__(authport, acctport, coaport, dict)

        self.hosts = hosts or {}
        self.auth_enabled = auth_enabled
        self.authfds: list[socket.socket] = []
        self.acct_enabled = acct_enabled
        self.acctfds: list = []
        self.coa_enabled = coa_enabled
        self.coafds: list = []

        if addresses:
            for addr in addresses:
                self.BindToAddress(addr)

    def _GetAddrInfo(
        self, addr: str
    ) -> set[tuple[socket.AddressFamily, str | int]] | list:
        """Use getaddrinfo to lookup all addresses for each address.

        Returns a list of tuples or an empty list:
          [(family, address)]

        Args:
            adddr (str): IP address to lookup
        """
        results = set()
        try:
            tmp = socket.getaddrinfo(addr, 80)
        except socket.gaierror:
            return []

        for el in tmp:
            results.add((el[0], el[4][0]))

        return results

    def BindToAddress(self, addr: str) -> None:
        """Add an address to listen on a specific interface.
        String "0.0.0.0" indicates you want to listen on all interfaces.

        Args:
            addr (str): IP address to listen on
        """
        addrFamily = self._GetAddrInfo(addr)
        for family, address in addrFamily:
            if self.auth_enabled:
                authfd = socket.socket(family, socket.SOCK_DGRAM)
                authfd.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
                authfd.bind((address, self.authport))
                self.authfds.append(authfd)

            if self.acct_enabled:
                acctfd = socket.socket(family, socket.SOCK_DGRAM)
                acctfd.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
                acctfd.bind((address, self.acctport))
                self.acctfds.append(acctfd)

            if self.coa_enabled:
                coafd = socket.socket(family, socket.SOCK_DGRAM)
                coafd.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
                coafd.bind((address, self.coaport))
                self.coafds.append(coafd)

    def HandleAuthPacket(self, pkt: packet.Packet):
        """Authentication packet handler.
        This is an empty function that is called when a valid
        authentication packet has been received. It can be overriden in
        derived classes to add custom behaviour.

        Args:
            pkt (packet.Packet): Packet to process
        """

    def HandleAcctPacket(self, pkt: packet.Packet):
        """Accounting packet handler.
        This is an empty function that is called when a valid
        accounting packet has been received. It can be overriden in
        derived classes to add custom behaviour.

        Args:
            pkt (packet.Packet): Packet to process
        """

    def HandleCoaPacket(self, pkt: packet.Packet):
        """CoA packet handler.
        This is an empty function that is called when a valid
        accounting packet has been received. It can be overriden in
        derived classes to add custom behaviour.

        Args:
            pkt (packet.Packet): Packet to process
        """

    def HandleDisconnectPacket(self, pkt: packet.Packet):
        """CoA packet handler.
        This is an empty function that is called when a valid
        accounting packet has been received. It can be overriden in
        derived classes to add custom behaviour.

        Args:
            pkt (packet.Packet): Packet to process
        """

    def _AddSecret(self, pkt: packet.Packet) -> None:
        """Add secret to packets received and raise ServerPacketError
        for unknown hosts.

        Args:
            pkt (packet.Packet): Packet to process
        """
        if pkt.source[0] in self.hosts:
            pkt.secret = self.hosts[pkt.source[0]].secret
        elif "0.0.0.0" in self.hosts:
            pkt.secret = self.hosts["0.0.0.0"].secret
        else:
            raise ServerPacketError("Received packet from unknown host")

    def _HandleAuthPacket(self, pkt: packet.Packet) -> None:
        """Process a packet received on the authentication port.
        If this packet should be dropped instead of processed a
        ServerPacketError exception should be raised. The main loop will
        drop the packet and log the reason.

        Args:
            pkt (packet.Packet): Packet to process
        """
        self._AddSecret(pkt)
        if pkt.code != PacketType.AccessRequest:
            raise ServerPacketError(
                "Received non-authentication packet on authentication port"
            )
        self.HandleAuthPacket(pkt)

    def _HandleAcctPacket(self, pkt: packet.Packet) -> None:
        """Process a packet received on the accounting port.
        If this packet should be dropped instead of processed a
        ServerPacketError exception should be raised. The main loop will
        drop the packet and log the reason.

        Args:
            pkt (packet.Packet): Packet to process
        """
        self._AddSecret(pkt)
        if pkt.code not in [
            PacketType.AccountingRequest,
            PacketType.AccountingResponse,
        ]:
            raise ServerPacketError("Received non-accounting packet on accounting port")
        self.HandleAcctPacket(pkt)

    def _HandleCoaPacket(self, pkt: packet.Packet) -> None:
        """Process a packet received on the coa port.
        If this packet should be dropped instead of processed a
        ServerPacketError exception should be raised. The main loop will
        drop the packet and log the reason.

        Args:
            pkt (packet.Packet): Packet to process
        """
        self._AddSecret(pkt)
        if pkt.code == PacketType.CoARequest:
            self.HandleCoaPacket(pkt)
        elif pkt.code == PacketType.DisconnectRequest:
            self.HandleDisconnectPacket(pkt)
        else:
            raise ServerPacketError("Received non-coa packet on coa port")

    def _GrabPacket(self, pktgen: Callable, fd: socket.socket) -> packet.Packet:
        """Read a packet from a network connection.
        This method assumes there is data waiting for to be read.

        Args:
            fd (socket.socket): Socket to read packet from

        Returns:
            packet.Packet: RADIUS packet
        """
        (data, source) = fd.recvfrom(self.MaxPacketSize)
        pkt = pktgen(data)
        pkt.source = source
        pkt.fd = fd
        return pkt

    def _PrepareSockets(self) -> None:
        """Prepare all sockets to receive packets."""
        for fd in self.authfds + self.acctfds + self.coafds:
            self._fdmap[fd.fileno()] = fd
            if os.name == 'nt':
                self._sel.register(fd.fileno(), selectors.EVENT_READ)
            else:
                self._poll.register(
                    fd.fileno(), select.POLLIN | select.POLLPRI | select.POLLERR
                )
        if self.auth_enabled:
            self._realauthfds = list(map(lambda x: x.fileno(), self.authfds))
        if self.acct_enabled:
            self._realacctfds = list(map(lambda x: x.fileno(), self.acctfds))
        if self.coa_enabled:
            self._realcoafds = list(map(lambda x: x.fileno(), self.coafds))

    def CreateReplyPacket(self, pkt: packet.Packet, **attributes) -> packet.Packet:
        """Create a reply packet.
        Create a new packet which can be returned as a reply to a received
        packet.

        Args:
            pkt (packet.Packet): Packet to process
        """
        reply = pkt.CreateReply(**attributes)
        reply.source = pkt.source
        return reply

    def _ProcessInput(self, fd: socket.socket) -> None:
        """Process available data.
        If this packet should be dropped instead of processed a
        PacketError exception should be raised. The main loop will
        drop the packet and log the reason.

        This function calls either HandleAuthPacket() or
        HandleAcctPacket() depending on which socket is being
        processed.

        Args:
            fd (socket.socket): Socket to read the packet from
        """
        if self.auth_enabled and fd.fileno() in self._realauthfds:
            pkt = self._GrabPacket(
                lambda data, s=self: s.CreateAuthPacket(packet=data), fd
            )
            self._HandleAuthPacket(pkt)
        elif self.acct_enabled and fd.fileno() in self._realacctfds:
            pkt = self._GrabPacket(
                lambda data, s=self: s.CreateAcctPacket(packet=data), fd
            )
            self._HandleAcctPacket(pkt)
        elif self.coa_enabled:
            pkt = self._GrabPacket(
                lambda data, s=self: s.CreateCoAPacket(packet=data), fd
            )
            self._HandleCoaPacket(pkt)
        else:
            raise ServerPacketError("Received packet for unknown handler")

    def Run(self) -> None:
        """Main loop.
        This method is the main loop for a RADIUS server. It waits
        for packets to arrive via the network and calls other methods
        to process them.
        """
        if os.name == 'nt':
            self._sel = selectors.DefaultSelector()
        else:
            self._poll = select.poll()
        self._fdmap: dict[int, socket.socket] = {}
        self._PrepareSockets()

        while True:
            if os.name == 'nt':
                for key, mask in self._sel.select(timeout=1):
                    if mask & selectors.EVENT_READ:
                        try:
                            fdo = self._fdmap[key.fd]
                            self._ProcessInput(fdo)
                        except ServerPacketError as err:
                            logger.info("Dropping packet: " + str(err))
                        except packet.PacketError as err:
                            logger.info("Received a broken packet: " + str(err))
                    else:
                        logger.error("Unexpected event in server main loop")
            else:
                for fd, event in self._poll.poll():
                    if event == select.POLLIN:
                        try:
                            fdo = self._fdmap[fd]
                            self._ProcessInput(fdo)
                        except ServerPacketError as err:
                            logger.info("Dropping packet: " + str(err))
                        except packet.PacketError as err:
                            logger.info("Received a broken packet: " + str(err))
                    else:
                        logger.error("Unexpected event in server main loop")



# Packet codes
from enum import IntEnum


class PacketType(IntEnum):
    AccessRequest = 1
    AccessAccept = 2
    AccessReject = 3
    AccountingRequest = 4
    AccountingResponse = 5
    AccessChallenge = 11
    StatusServer = 12
    StatusClient = 13
    DisconnectRequest = 40
    DisconnectACK = 41
    DisconnectNAK = 42
    CoARequest = 43
    CoAACK = 44
    CoANAK = 45


class EAPPacketType(IntEnum):
    REQUEST = 1
    RESPONSE = 2


class EAPType(IntEnum):
    IDENTITY = 1


DATATYPES = frozenset(
    [
        "string",
        "ipaddr",
        "integer",
        "date",
        "octets",
        "abinary",
        "ipv6addr",
        "ipv6prefix",
        "short",
        "byte",
        "signed",
        "ifid",
        "ether",
        "tlv",
        "integer64",
    ]
)



import binascii
import ssl
import struct
from asyncio import StreamReader
from collections.abc import Buffer
from hashlib import sha256
from ipaddress import IPv4Address, IPv4Network, IPv6Address, IPv6Network


def EncodeString(origstr: str) -> bytes:
    """Encode a string to bytes, ensuring it is UTF-8 encoded."""
    if len(origstr) > 253:
        raise ValueError("Can only encode strings of <= 253 characters")
    if isinstance(origstr, str):
        return origstr.encode("utf-8")
    else:
        return origstr


def EncodeOctets(octetstring: str) -> str | bytes:
    """Encode raw octet string (already in bytes)."""
    # Check for max length of the hex encoded with 0x prefix, as a sanity check
    if len(octetstring) > 508:
        raise ValueError("Can only encode strings of <= 253 characters")

    hexstring: str | bytes
    encoded_octets: str | bytes
    if isinstance(octetstring, bytes) and octetstring.startswith(b"0x"):
        hexstring = octetstring.split(b"0x")[1]
        encoded_octets = binascii.unhexlify(hexstring)
    elif isinstance(octetstring, str) and octetstring.startswith("0x"):
        hexstring = octetstring.split("0x")[1]
        encoded_octets = binascii.unhexlify(hexstring)
    elif isinstance(octetstring, str) and octetstring.isdecimal():
        encoded_octets = struct.pack(">L", int(octetstring)).lstrip(b"\x00")
    else:
        encoded_octets = octetstring

    # Check for the encoded value being longer than 253 chars
    if len(encoded_octets) > 253:
        raise ValueError("Can only encode strings of <= 253 characters")

    return encoded_octets


def EncodeAddress(addr: str) -> bytes:
    """Encode an IPv4 address (dotted string) to 4-byte format."""
    if not isinstance(addr, str):
        raise TypeError("Address has to be a string")
    return IPv4Address(addr).packed


def EncodeIPv6Prefix(addr: str) -> bytes:
    """Encode an IPv6 address and prefix length to 18-byte format."""
    if not isinstance(addr, str):
        raise TypeError("IPv6 Prefix has to be a string")
    ip = IPv6Network(addr, strict=False)
    return struct.pack("2B", *[0, ip.prefixlen]) + ip.network_address.packed


def EncodeIPv6Address(addr: str) -> bytes:
    """Encode an IPv6 address (as string) to 16-byte format."""
    if not isinstance(addr, str):
        raise TypeError("IPv6 Address has to be a string")
    return IPv6Address(addr).packed


def EncodeAscendBinary(orig_str: str) -> bytes:
    """Encode binary data in Ascend-specific format (length prefixed)."""
    """
    Format: List of type=value pairs separated by spaces.

    Example: 'family=ipv4 action=discard direction=in dst=10.10.255.254/32'

    Note: redirect(0x20) action is added for http-redirect (walled garden) use case

    Type:
        family      ipv4(default) or ipv6
        action      discard(default) or accept or redirect
        direction   in(default) or out
        src         source prefix (default ignore)
        dst         destination prefix (default ignore)
        proto       protocol number / next-header number (default ignore)
        sport       source port (default ignore)
        dport       destination port (default ignore)
        sportq      source port qualifier (default 0)
        dportq      destination port qualifier (default 0)

    Source/Destination Port Qualifier:
        0   no compare
        1   less than
        2   equal to
        3   greater than
        4   not equal to
    """

    terms = {
        "family": b"\x01",
        "action": b"\x00",
        "direction": b"\x01",
        "src": b"\x00\x00\x00\x00",
        "dst": b"\x00\x00\x00\x00",
        "srcl": b"\x00",
        "dstl": b"\x00",
        "proto": b"\x00",
        "sport": b"\x00\x00",
        "dport": b"\x00\x00",
        "sportq": b"\x00",
        "dportq": b"\x00",
    }

    family = "ipv4"
    ip: IPv4Network | IPv6Network

    for t in orig_str.split(" "):
        key, value = t.split("=")
        if key == "family" and value == "ipv6":
            family = "ipv6"
            terms[key] = b"\x03"
            if terms["src"] == b"\x00\x00\x00\x00":
                terms["src"] = 16 * b"\x00"
            if terms["dst"] == b"\x00\x00\x00\x00":
                terms["dst"] = 16 * b"\x00"
        elif key == "action" and value == "accept":
            terms[key] = b"\x01"
        elif key == "action" and value == "redirect":
            terms[key] = b"\x20"
        elif key == "direction" and value == "out":
            terms[key] = b"\x00"
        elif key == "src" or key == "dst":
            if family == "ipv4":
                ip = IPv4Network(value)
            else:
                ip = IPv6Network(value)
            terms[key] = ip.network_address.packed
            terms[key + "l"] = struct.pack("B", ip.prefixlen)
        elif key == "sport" or key == "dport":
            terms[key] = struct.pack("!H", int(value))
        elif key == "sportq" or key == "dportq" or key == "proto":
            terms[key] = struct.pack("B", int(value))

    trailer = 8 * b"\x00"

    result = b"".join(
        (
            terms["family"],
            terms["action"],
            terms["direction"],
            b"\x00",
            terms["src"],
            terms["dst"],
            terms["srcl"],
            terms["dstl"],
            terms["proto"],
            b"\x00",
            terms["sport"],
            terms["dport"],
            terms["sportq"],
            terms["dportq"],
            b"\x00\x00",
            trailer,
        )
    )
    return result


def EncodeInteger(num: int, format: str = "!I") -> bytes:
    """Encode a 32-bit unsigned integer to 4-byte big-endian."""
    try:
        num = int(num)
    except (ValueError, TypeError):
        raise TypeError("Can not encode non-integer as integer")
    return struct.pack(format, num)


def EncodeInteger64(num: int, format: str = "!Q") -> bytes:
    """Encode a 64-bit unsigned integer to 8-byte big-endian."""
    try:
        num = int(num)
    except (ValueError, TypeError):
        raise TypeError("Can not encode non-integer as integer64")
    return struct.pack(format, num)


def EncodeDate(num: int) -> bytes:
    """Encode a UNIX timestamp (int) to 4-byte format."""
    if not isinstance(num, int):
        raise TypeError("Can not encode non-integer as date")
    return struct.pack("!I", num)


def DecodeString(orig_str: bytes) -> str:
    """Decode UTF-8 bytes into a string."""
    try:
        return orig_str.decode("utf-8")
    except UnicodeDecodeError:
        # Non-UTF-8 data displayed in hexadecimal form
        return orig_str.hex()


def DecodeOctets(orig_bytes: bytes) -> bytes:
    """Return bytes unchanged (octet format)."""
    return orig_bytes


def DecodeAddress(addr: Buffer) -> str:
    """Decode 4-byte data into an IPv4 dotted string."""
    return ".".join(map(str, struct.unpack("BBBB", addr)))


def DecodeIPv6Prefix(addr: bytes | bytearray) -> str:
    """Decode 18-byte IPv6 prefix format into address/prefix tuple."""
    addr = addr + b"\x00" * (18 - len(addr))
    _, length, prefix = ":".join(
        map("{:x}".format, struct.unpack("!BB" + "H" * 8, addr))
    ).split(":", 2)
    return str(IPv6Network("{}/{}".format(prefix, int(length, 16))))


def DecodeIPv6Address(addr: bytes | bytearray) -> str:
    """Decode 16-byte IPv6 address into a readable string."""
    addr = addr + b"\x00" * (16 - len(addr))
    prefix = ":".join(map("{:x}".format, struct.unpack("!" + "H" * 8, addr)))
    return str(IPv6Address(prefix))


def DecodeAscendBinary(orig_bytes: bytes) -> bytes:
    """Decode Ascend-specific binary format (length-prefixed)."""
    return orig_bytes


def DecodeInteger(num: Buffer, format: str = "!I") -> bytes:
    """Decode 4-byte big-endian unsigned integer."""
    return (struct.unpack(format, num))[0]


def DecodeInteger64(num: Buffer, format: str = "!Q") -> bytes:
    """Decode 8-byte big-endian unsigned integer."""
    return (struct.unpack(format, num))[0]


def DecodeDate(num: Buffer) -> bytes:
    """Decode 4-byte UNIX timestamp into an integer."""
    return (struct.unpack("!I", num))[0]


def EncodeAttr(datatype: str, value) -> bytes | str:
    """Encode a RADIUS attribute (type, value, length) into bytes."""
    if datatype == "string":
        return EncodeString(value)
    elif datatype == "octets":
        return EncodeOctets(value)
    elif datatype == "integer":
        return EncodeInteger(value)
    elif datatype == "ipaddr":
        return EncodeAddress(value)
    elif datatype == "ipv6prefix":
        return EncodeIPv6Prefix(value)
    elif datatype == "ipv6addr":
        return EncodeIPv6Address(value)
    elif datatype == "abinary":
        return EncodeAscendBinary(value)
    elif datatype == "signed":
        return EncodeInteger(value, "!i")
    elif datatype == "short":
        return EncodeInteger(value, "!H")
    elif datatype == "byte":
        return EncodeInteger(value, "!B")
    elif datatype == "date":
        return EncodeDate(value)
    elif datatype == "integer64":
        return EncodeInteger64(value)
    else:
        raise ValueError("Unknown attribute type %s" % datatype)


def DecodeAttr(datatype: str, value) -> bytes | str:
    """Decode a RADIUS attribute from bytes into a type and value."""
    if datatype == "string":
        return DecodeString(value)
    elif datatype == "octets":
        return DecodeOctets(value)
    elif datatype == "integer":
        return DecodeInteger(value)
    elif datatype == "ipaddr":
        return DecodeAddress(value)
    elif datatype == "ipv6prefix":
        return DecodeIPv6Prefix(value)
    elif datatype == "ipv6addr":
        return DecodeIPv6Address(value)
    elif datatype == "abinary":
        return DecodeAscendBinary(value)
    elif datatype == "signed":
        return DecodeInteger(value, "!i")
    elif datatype == "short":
        return DecodeInteger(value, "!H")
    elif datatype == "byte":
        return DecodeInteger(value, "!B")
    elif datatype == "date":
        return DecodeDate(value)
    elif datatype == "integer64":
        return DecodeInteger64(value)
    else:
        raise ValueError("Unknown attribute type %s" % datatype)


def get_cert_fingerprint(cert: bytes) -> str:
    """Generate SHA-256 fingerprint from a certificate."""
    der_bytes = ssl.PEM_cert_to_DER_cert(ssl.DER_cert_to_PEM_cert(cert))
    hash = sha256(der_bytes).digest()
    # Return in base64 or hex
    return hash.hex()  # or base64.b64encode(sha256).decode()


async def read_radius_packet(reader: StreamReader) -> bytes:
    """Read a full RADIUS packet from the stream.

    There's no built-in framing in RadSec, so we can't read a fixed-size packet.
    Instead, we read the header first to determine the length of the packet,
    and then read the rest of the packet based on that length.

    RADIUS packets are prefixed with a 4-byte header:
        - Code (1 byte)
        - Identifier (1 byte)
        - Length (2 bytes)

    The length includes the header, so the minimum length is 20 bytes
    (4-byte header + 16-byte Authenticator).
    If the length is less than 20, it is considered invalid.

    :param reader: asyncio StreamReader to read from
    :return: Full RADIUS packet as bytes
    """
    header = await reader.readexactly(4)
    code, identifier, length = struct.unpack("!BBH", header)

    if length < 20:
        raise ValueError("Invalid RADIUS packet length")

    body = await reader.readexactly(length - 4)
    return header + body



import asyncio
import ssl
import struct
from hashlib import md5
from typing import Optional

from loguru import logger

from pyrad2.constants import EAPPacketType, EAPType, PacketType
from pyrad2.packet import (
    AcctPacket,
    AuthPacket,
    CoAPacket,
    CurrentID,
    Packet,
    PacketError,
    PacketImplementation,
)
from pyrad2.tools import read_radius_packet


class RadSecClient:
    def __init__(
        self,
        server: str = "127.0.0.1",
        port: int = 2083,
        secret: bytes = b"radsec",
        dict=None,
        retries: int = 3,
        timeout: int = 5,
        certfile: str = "certs/client/client.cert.pem",
        keyfile: str = "certs/client/client.key.pem",
        certfile_server: str = "certs//ca/ca.cert.pem",
        check_hostname: bool = False,
    ):
        """Initializes a RadSec client.

        Args:
            server (str): IP address to connect to.
            port (int): RadSec port, defaults to 2083.
            secret (bytes): Secret. Defaults to radsec as per RFC 6614.
                Different implementations support setting an arbitrary
                shared secret but if you want to stick to the RFC,
                the shared secret must be `radsec`.
            dict (Dictionary): RADIUS dictionary to use.
            certfile (str): Path to client SSL certificate
            keyfile (str): Path to client SSL certificate
            certfile_server (str): Path to server SSL certificate

        """
        self.server = server
        self.port = port
        self.secret = secret
        self.retries = retries
        self.timeout = timeout
        self.dict = dict

        self.setup_ssl(certfile, keyfile, certfile_server, check_hostname)

    def setup_ssl(
        self, certfile: str, keyfile: str, certfile_server: str, check_hostname: bool
    ):
        try:
            self.ssl_ctx = ssl.create_default_context(
                ssl.Purpose.SERVER_AUTH, cafile=certfile_server
            )

            self.ssl_ctx.load_cert_chain(certfile=certfile, keyfile=keyfile)
        except FileNotFoundError as e:
            ssl_paths = ", ".join([certfile, keyfile, certfile_server])
            msg = "One or more SSL files could not be found. Current paths: {}"
            logger.error(msg, ssl_paths)
            raise FileNotFoundError(msg.format(ssl_paths)) from e

        self.ssl_ctx.check_hostname = check_hostname

    def create_auth_packet(self, **kwargs) -> AuthPacket:
        """Create a new RADIUS packet.
        This utility function creates a new RADIUS packet which can
        be used to communicate with the RADIUS server this client
        talks to. This is initializing the new packet with the
        dictionary and secret used for the client.

        Returns:
            Packet: A new AuthPacket instance
        """
        id = kwargs.pop("id", Packet.CreateID())
        return AuthPacket(
            dict=self.dict,
            id=id,
            secret=self.secret,
            **kwargs,
        )

    def create_acct_packet(self, **kwargs) -> AcctPacket:
        """Create a new RADIUS packet.
        This utility function creates a new RADIUS packet which can
        be used to communicate with the RADIUS server this client
        talks to. This is initializing the new packet with the
        dictionary and secret used for the client.

        Returns:
            Packet: A new AcctPacket instance
        """
        id = kwargs.pop("id", Packet.CreateID())
        return AcctPacket(
            id=id,
            dict=self.dict,
            secret=self.secret,
            **kwargs,
        )

    def create_coa_packet(self, **kwargs) -> CoAPacket:
        """Create a new RADIUS packet.
        This utility function creates a new RADIUS packet which can
        be used to communicate with the RADIUS server this client
        talks to. This is initializing the new packet with the
        dictionary and secret used for the client.

        Returns:
            Packet: A new CoA packet instance
        """
        id = kwargs.pop("id", Packet.CreateID())
        return CoAPacket(id=id, dict=self.dict, secret=self.secret, **kwargs)

    def create_packet(self, id, **kwargs) -> Packet:
        return Packet(id=id, dict=self.dict, secret=self.secret, **kwargs)

    async def _send_packet(self, packet: PacketImplementation) -> Optional[Packet]:
        """Send a packet to a RADIUS server.

        Args:
            packet (Packet): The packet to send
        """
        reader, writer = await asyncio.open_connection(
            self.server, self.port, ssl=self.ssl_ctx
        )

        logger.info(
            "Connected to RADSEC server on {}:{}, sending RADIUS packet",
            self.server,
            self.port,
        )

        writer.write(packet.RequestPacket())
        await writer.drain()

        async def close():
            writer.close()
            await writer.wait_closed()

        try:
            response = await read_radius_packet(reader)
        except asyncio.TimeoutError:
            logger.warning("Timeout waiting for server response")
            await close()
            return None

        if not response:
            logger.info("No response received")
            await close()
            return None

        logger.info("Received {} bytes from server", len(response))
        logger.debug("Response: {}", response.hex())

        try:
            reply = packet.CreateReply(packet=response)
            if packet.VerifyReply(reply, response):
                return reply
        except PacketError as e:
            logger.error("Error creating reply {}", e)
            await close()

        return None

    async def send_packet(self, packet: PacketImplementation) -> Optional[Packet]:
        """Send a packet to a RADIUS server.

        Args:
            packet (Packet): The packet to send
        """
        if isinstance(packet, AuthPacket):
            if packet.auth_type == "eap-md5":
                # Creating EAP-Identity
                password = packet[2][0] if 2 in packet else packet[1][0]
                packet[79] = [
                    struct.pack(
                        "!BBHB%ds" % len(password),
                        EAPPacketType.RESPONSE,
                        CurrentID,
                        len(password) + 5,
                        EAPType.IDENTITY,
                        password,
                    )
                ]
            reply = await self._send_packet(packet)
            if (
                reply
                and reply.code == PacketType.AccessChallenge
                and packet.auth_type == "eap-md5"
            ):
                # Got an Access-Challenge
                eap_code, eap_id, eap_size, eap_type, eap_md5 = struct.unpack(
                    "!BBHB%ds" % (len(reply[79][0]) - 5), reply[79][0]
                )
                # Sending back an EAP-Type-MD5-Challenge
                # Thank god for http://www.secdev.org/python/eapy.py
                client_pw = packet[2][0] if 2 in packet else packet[1][0]
                md5_challenge = md5(
                    struct.pack("!B", eap_id) + client_pw + eap_md5[1:]
                ).digest()
                packet[79] = [
                    struct.pack(
                        "!BBHBB",
                        2,
                        eap_id,
                        len(md5_challenge) + 6,
                        4,
                        len(md5_challenge),
                    )
                    + md5_challenge
                ]
                # Copy over Challenge-State
                packet[24] = reply[24]
                reply = await self._send_packet(packet)
            return reply
        elif isinstance(packet, CoAPacket):
            return await self._send_packet(packet)
        else:
            return await self._send_packet(packet)



import os
import unittest

from pyrad2.constants import PacketType
from pyrad2.dictionary import Dictionary
from pyrad2.radsec.client import RadSecClient
from pyrad2.radsec.server import RadSecServer as BaseRadSecServer
from pyrad2.radsec.server import UnknownHost
from pyrad2.server import RemoteHost

from .base import TEST_ROOT_PATH

TEST_HOST = RemoteHost(
    "name",
    b"radsec",
    "127.0.0.1",
)

SERVER_CERTFILE = os.path.join(TEST_ROOT_PATH, "certs/server/server.cert.pem")
SERVER_KEYFILE = os.path.join(TEST_ROOT_PATH, "certs/server/server.key.pem")
CA_CERTFILE = os.path.join(TEST_ROOT_PATH, "certs/ca/ca.cert.pem")
CLIENT_CERTFILE = os.path.join(TEST_ROOT_PATH, "certs/client/client.cert.pem")
CLIENT_KEYFILE = os.path.join(TEST_ROOT_PATH, "certs/client/client.key.pem")


class RemoteHostTests(unittest.TestCase):
    def test_simple_construction(self):
        host = RemoteHost(
            "127.0.0.1",
            b"radsec",
            "name",
        )
        self.assertEqual(host.name, "name")
        self.assertEqual(host.address, "127.0.0.1")
        self.assertEqual(host.secret, b"radsec")


class RadSecServer(BaseRadSecServer):
    async def handle_access_request(self, packet):
        reply = packet.CreateReply(
            **{
                "Service-Type": "Framed-User",
                "Framed-IP-Address": "192.168.0.1",
                "Framed-IPv6-Prefix": "fc66::1/64",
            },
        )

        reply.code = PacketType.AccessAccept
        return reply

    async def handle_accounting(self, packet):
        return packet.CreateReply()

    async def handle_disconnect(self, packet):
        reply = packet.CreateReply()
        reply.code = 45  # COA NAK
        return reply

    async def handle_coa(self, packet):
        return packet.CreateReply()


class ServerTests(unittest.IsolatedAsyncioTestCase):
    def setUp(self):
        self.dictionary = Dictionary(os.path.join(TEST_ROOT_PATH, "dicts/dictionary"))

        self.server = RadSecServer(
            certfile=SERVER_CERTFILE,
            keyfile=SERVER_KEYFILE,
            ca_certfile=CA_CERTFILE,
            dictionary=self.dictionary,
        )
        self.server.hosts = {"127.0.0.1": TEST_HOST}

        self.client = RadSecClient(
            server="127.0.0.1",
            secret=b"radsec",
            dict=self.dictionary,
            certfile=CLIENT_CERTFILE,
            keyfile=CLIENT_KEYFILE,
            certfile_server=CA_CERTFILE,
        )

    def test_simple_construction(self):
        self.assertEqual(self.server.listen_address, "0.0.0.0")
        self.assertEqual(self.server.listen_port, 2083)
        self.assertEqual(self.server.hosts, {"127.0.0.1": TEST_HOST})
        self.assertEqual(self.server.dict, self.dictionary)
        self.assertEqual(self.server.verify_packet, False)

    async def test_unknown_host(self):
        with self.assertRaises(UnknownHost):
            await self.server.packet_received({}, "4.4.4.4")


class AuthPacketHandlingTests(ServerTests):
    def setUp(self):
        super().setUp()
        self.packet = self.create_auth_packet()

    def create_auth_packet(self):
        packet = self.client.create_auth_packet(
            code=PacketType.AccessRequest, User_Name="wichert"
        )
        packet["NAS-IP-Address"] = "192.168.1.10"
        packet["NAS-Port"] = 0
        packet["Service-Type"] = "Login-User"
        packet["NAS-Identifier"] = "trillian"
        packet["Called-Station-Id"] = "00-04-5F-00-0F-D1"
        packet["Calling-Station-Id"] = "00-01-24-80-B3-9C"
        packet["Framed-IP-Address"] = "10.0.0.100"
        return packet

    async def test_handle_auth_packet(self):
        reply = await self.server.handle_access_request(self.packet)
        self.assertEqual(reply.code, PacketType.AccessAccept)


class AcctPacketHandlingTests(ServerTests):
    def setUp(self):
        super().setUp()
        self.packet = self.create_acct_packet()

    def create_acct_packet(self):
        packet = self.client.create_acct_packet(
            code=PacketType.AccountingRequest, User_Name="wichert"
        )
        packet["NAS-IP-Address"] = "192.168.1.10"
        packet["NAS-Port"] = 0
        packet["Service-Type"] = "Login-User"
        packet["NAS-Identifier"] = "trillian"
        packet["Called-Station-Id"] = "00-04-5F-00-0F-D1"
        packet["Calling-Station-Id"] = "00-01-24-80-B3-9C"
        packet["Framed-IP-Address"] = "10.0.0.100"
        packet["Acct-Status-Type"] = "Start"
        return packet

    async def test_handle_acct_packet(self):
        reply = await self.server.handle_accounting(self.packet)
        self.assertEqual(reply.code, PacketType.AccountingResponse)



from typing import Any, Dict, Hashable


class BiDict:
    """BiDict (Bidirectional Dictionary) provides a one-to-one mapping
    between keys and values.

    Supports both forward and reverse lookup.
    """

    def __init__(self) -> None:
        """Initialize empty forward and reverse dictionaries."""
        self.forward: Dict[Hashable, Any] = {}
        self.backward: Dict[Hashable, Any] = {}

    def Add(self, one: Hashable, two: Hashable) -> None:
        """Add a bidirectional mapping between 'one' and 'two'."""
        self.forward[one] = two
        self.backward[two] = one

    def __len__(self) -> int:
        """Return the number of entries in the dictionary."""
        return len(self.forward)

    def __getitem__(self, key: Hashable) -> Any:
        """Retrieve the value associated with the given key."""
        return self.GetForward(key)

    def __delitem__(self, key: Hashable) -> None:
        """Remove key and its associated value from the dictionary."""
        if key in self.forward:
            del self.backward[self.forward[key]]
            del self.forward[key]
        else:
            del self.forward[self.backward[key]]
            del self.backward[key]

    def GetForward(self, key: Hashable) -> Any:
        """Return the value associated with 'key' from the forward mapping."""
        return self.forward[key]

    def HasForward(self, key: Hashable) -> bool:
        """Check if 'key' exists in the forward mapping."""
        return key in self.forward

    def GetBackward(self, key: Hashable) -> Any:
        """Return the key associated with 'value' from the reverse mapping."""
        return self.backward[key]

    def HasBackward(self, key: Hashable) -> bool:
        """Check if 'value' exists in the reverse mapping."""
        return key in self.backward



import operator
import os
import unittest
from io import StringIO

from pyrad2.dictfile import DictFile
from pyrad2.dictionary import Attribute, Dictionary, ParseError
from pyrad2.tools import DecodeAttr

from .base import TEST_ROOT_PATH


class AttributeTests(unittest.TestCase):
    def testInvalidDataType(self):
        self.assertRaises(ValueError, Attribute, "name", "code", "datatype")

    def testConstructionParameters(self):
        attr = Attribute("name", "code", "integer", False, "vendor")
        self.assertEqual(attr.name, "name")
        self.assertEqual(attr.code, "code")
        self.assertEqual(attr.type, "integer")
        self.assertEqual(attr.is_sub_attribute, False)
        self.assertEqual(attr.vendor, "vendor")
        self.assertEqual(len(attr.values), 0)
        self.assertEqual(len(attr.sub_attributes), 0)

    def testNamedConstructionParameters(self):
        attr = Attribute(name="name", code="code", datatype="integer", vendor="vendor")
        self.assertEqual(attr.name, "name")
        self.assertEqual(attr.code, "code")
        self.assertEqual(attr.type, "integer")
        self.assertEqual(attr.vendor, "vendor")
        self.assertEqual(len(attr.values), 0)

    def testValues(self):
        attr = Attribute(
            "name",
            "code",
            "integer",
            False,
            "vendor",
            dict(pie="custard", shake="vanilla"),
        )
        self.assertEqual(len(attr.values), 2)
        self.assertEqual(attr.values["shake"], "vanilla")


class DictionaryInterfaceTests(unittest.TestCase):
    def testEmptyDictionary(self):
        dict = Dictionary()
        self.assertEqual(len(dict), 0)

    def testContainment(self):
        dict = Dictionary()
        self.assertEqual("test" in dict, False)
        self.assertEqual(dict.has_key("test"), False)
        dict.attributes["test"] = "dummy"
        self.assertEqual("test" in dict, True)
        self.assertEqual(dict.has_key("test"), True)

    def testReadonlyContainer(self):
        dict = Dictionary()
        self.assertRaises(TypeError, operator.setitem, dict, "test", "dummy")
        self.assertRaises(AttributeError, operator.attrgetter("clear"), dict)
        self.assertRaises(AttributeError, operator.attrgetter("update"), dict)


class DictionaryParsingTests(unittest.TestCase):
    simple_dict_values = [
        ("Test-String", 1, "string"),
        ("Test-Octets", 2, "octets"),
        ("Test-Integer", 0x03, "integer"),
        ("Test-Ip-Address", 4, "ipaddr"),
        ("Test-Ipv6-Address", 5, "ipv6addr"),
        ("Test-If-Id", 6, "ifid"),
        ("Test-Date", 7, "date"),
        ("Test-Abinary", 8, "abinary"),
        ("Test-Tlv", 9, "tlv"),
        ("Test-Tlv-Str", 1, "string"),
        ("Test-Tlv-Int", 2, "integer"),
        ("Test-Integer64", 10, "integer64"),
        ("Test-Integer64-Hex", 10, "integer64"),
        ("Test-Integer64-Oct", 10, "integer64"),
    ]

    def setUp(self):
        self.path = os.path.join(TEST_ROOT_PATH, "data")
        self.dict = Dictionary(os.path.join(self.path, "simple"))

    def testParseEmptyDictionary(self):
        dict = Dictionary(StringIO(""))
        self.assertEqual(len(dict), 0)

    def testParseMultipleDictionaries(self):
        dict = Dictionary(StringIO(""))
        self.assertEqual(len(dict), 0)
        one = StringIO("ATTRIBUTE Test-First 1 string")
        two = StringIO("ATTRIBUTE Test-Second 2 string")
        dict = Dictionary(StringIO(""), one, two)
        self.assertEqual(len(dict), 2)

    def testParseSimpleDictionary(self):
        self.assertEqual(len(self.dict), len(self.simple_dict_values))
        for attr, code, type in self.simple_dict_values:
            attr = self.dict[attr]
            self.assertEqual(attr.code, code)
            self.assertEqual(attr.type, type)

    def testAttributeTooFewColumnsError(self):
        try:
            self.dict.ReadDictionary(StringIO("ATTRIBUTE Oops-Too-Few-Columns"))
        except ParseError as e:
            self.assertEqual("attribute" in str(e), True)
        else:
            self.fail()

    def testAttributeUnknownTypeError(self):
        try:
            self.dict.ReadDictionary(StringIO("ATTRIBUTE Test-Type 1 dummy"))
        except ParseError as e:
            self.assertEqual("dummy" in str(e), True)
        else:
            self.fail()

    def testAttributeUnknownVendorError(self):
        try:
            self.dict.ReadDictionary(StringIO("ATTRIBUTE Test-Type 1 Simplon"))
        except ParseError as e:
            self.assertEqual("Simplon" in str(e), True)
        else:
            self.fail()

    def testAttributeOptions(self):
        self.dict.ReadDictionary(
            StringIO("ATTRIBUTE Option-Type 1 string has_tag,encrypt=1")
        )
        self.assertEqual(self.dict["Option-Type"].has_tag, True)
        self.assertEqual(self.dict["Option-Type"].encrypt, 1)

    def testAttributeEncryptionError(self):
        try:
            self.dict.ReadDictionary(StringIO("ATTRIBUTE Test-Type 1 string encrypt=4"))
        except ParseError as e:
            self.assertEqual("encrypt" in str(e), True)
        else:
            self.fail()

    def testValueTooFewColumnsError(self):
        try:
            self.dict.ReadDictionary(StringIO("VALUE Oops-Too-Few-Columns"))
        except ParseError as e:
            self.assertEqual("value" in str(e), True)
        else:
            self.fail()

    def testValueForUnknownAttributeError(self):
        try:
            self.dict.ReadDictionary(StringIO("VALUE Test-Attribute Test-Text 1"))
        except ParseError as e:
            self.assertEqual("unknown attribute" in str(e), True)
        else:
            self.fail()

    def testIntegerValueParsing(self):
        self.assertEqual(len(self.dict["Test-Integer"].values), 0)
        self.dict.ReadDictionary(StringIO("VALUE Test-Integer Value-Six 5"))
        self.assertEqual(len(self.dict["Test-Integer"].values), 1)
        self.assertEqual(
            DecodeAttr("integer", self.dict["Test-Integer"].values["Value-Six"]), 5
        )

    def testInteger64ValueParsing(self):
        self.assertEqual(len(self.dict["Test-Integer64"].values), 0)
        self.dict.ReadDictionary(StringIO("VALUE Test-Integer64 Value-Six 5"))
        self.assertEqual(len(self.dict["Test-Integer64"].values), 1)
        self.assertEqual(
            DecodeAttr("integer64", self.dict["Test-Integer64"].values["Value-Six"]), 5
        )

    def testStringValueParsing(self):
        self.assertEqual(len(self.dict["Test-String"].values), 0)
        self.dict.ReadDictionary(StringIO("VALUE Test-String Value-Custard custardpie"))
        self.assertEqual(len(self.dict["Test-String"].values), 1)
        self.assertEqual(
            DecodeAttr("string", self.dict["Test-String"].values["Value-Custard"]),
            "custardpie",
        )

    def testOctetValueParsing(self):
        self.assertEqual(len(self.dict["Test-Octets"].values), 0)
        self.dict.ReadDictionary(
            StringIO(
                "ATTRIBUTE Test-Octets 1 octets\n"
                "VALUE Test-Octets Value-A 65\n"  # "A"
                "VALUE Test-Octets Value-B 0x42\n"
            )
        )  # "B"
        self.assertEqual(len(self.dict["Test-Octets"].values), 2)
        self.assertEqual(
            DecodeAttr("octets", self.dict["Test-Octets"].values["Value-A"]), b"A"
        )
        self.assertEqual(
            DecodeAttr("octets", self.dict["Test-Octets"].values["Value-B"]), b"B"
        )

    def testTlvParsing(self):
        self.assertEqual(len(self.dict["Test-Tlv"].sub_attributes), 2)
        self.assertEqual(
            self.dict["Test-Tlv"].sub_attributes, {1: "Test-Tlv-Str", 2: "Test-Tlv-Int"}
        )

    def testSubTlvParsing(self):
        for attr, _, _ in self.simple_dict_values:
            if attr.startswith("Test-Tlv-"):
                self.assertEqual(self.dict[attr].is_sub_attribute, True)
                self.assertEqual(self.dict[attr].parent, self.dict["Test-Tlv"])
            else:
                self.assertEqual(self.dict[attr].is_sub_attribute, False)
                self.assertEqual(self.dict[attr].parent, None)

        # tlv with vendor
        full_dict = Dictionary(os.path.join(self.path, "full"))
        self.assertEqual(full_dict["Simplon-Tlv-Str"].is_sub_attribute, True)
        self.assertEqual(full_dict["Simplon-Tlv-Str"].parent, full_dict["Simplon-Tlv"])
        self.assertEqual(full_dict["Simplon-Tlv-Int"].is_sub_attribute, True)
        self.assertEqual(full_dict["Simplon-Tlv-Int"].parent, full_dict["Simplon-Tlv"])

    def testVenderTooFewColumnsError(self):
        try:
            self.dict.ReadDictionary(StringIO("VENDOR Simplon"))
        except ParseError as e:
            self.assertEqual("vendor" in str(e), True)
        else:
            self.fail()

    def testVendorParsing(self):
        self.assertRaises(
            ParseError,
            self.dict.ReadDictionary,
            StringIO("ATTRIBUTE Test-Type 1 integer Simplon"),
        )
        self.dict.ReadDictionary(StringIO("VENDOR Simplon 42"))
        self.assertEqual(self.dict.vendors["Simplon"], 42)
        self.dict.ReadDictionary(StringIO("ATTRIBUTE Test-Type 1 integer Simplon"))
        self.assertEqual(self.dict.attrindex["Test-Type"], (42, 1))

    def testVendorOptionError(self):
        self.assertRaises(
            ParseError,
            self.dict.ReadDictionary,
            StringIO("ATTRIBUTE Test-Type 1 integer Simplon"),
        )
        try:
            self.dict.ReadDictionary(StringIO("VENDOR Simplon 42 badoption"))
        except ParseError as e:
            self.assertEqual("option" in str(e), True)
        else:
            self.fail()

    def testVendorFormatError(self):
        self.assertRaises(
            ParseError,
            self.dict.ReadDictionary,
            StringIO("ATTRIBUTE Test-Type 1 integer Simplon"),
        )
        try:
            self.dict.ReadDictionary(StringIO("VENDOR Simplon 42 format=5,4"))
        except ParseError as e:
            self.assertEqual("format" in str(e), True)
        else:
            self.fail()

    def testVendorFormatSyntaxError(self):
        self.assertRaises(
            ParseError,
            self.dict.ReadDictionary,
            StringIO("ATTRIBUTE Test-Type 1 integer Simplon"),
        )
        try:
            self.dict.ReadDictionary(StringIO("VENDOR Simplon 42 format=a,1"))
        except ParseError as e:
            self.assertEqual("Syntax" in str(e), True)
        else:
            self.fail()

    def testBeginVendorTooFewColumns(self):
        try:
            self.dict.ReadDictionary(StringIO("BEGIN-VENDOR"))
        except ParseError as e:
            self.assertEqual("begin-vendor" in str(e), True)
        else:
            self.fail()

    def testBeginVendorUnknownVendor(self):
        try:
            self.dict.ReadDictionary(StringIO("BEGIN-VENDOR Simplon"))
        except ParseError as e:
            self.assertEqual("Simplon" in str(e), True)
        else:
            self.fail()

    def testBeginVendorParsing(self):
        self.dict.ReadDictionary(
            StringIO(
                "VENDOR Simplon 42\nBEGIN-VENDOR Simplon\nATTRIBUTE Test-Type 1 integer"
            )
        )
        self.assertEqual(self.dict.attrindex["Test-Type"], (42, 1))

    def testEndVendorUnknownVendor(self):
        try:
            self.dict.ReadDictionary(StringIO("END-VENDOR"))
        except ParseError as e:
            self.assertEqual("end-vendor" in str(e), True)
        else:
            self.fail()

    def testEndVendorUnbalanced(self):
        try:
            self.dict.ReadDictionary(
                StringIO("VENDOR Simplon 42\nBEGIN-VENDOR Simplon\nEND-VENDOR Oops\n")
            )
        except ParseError as e:
            self.assertEqual("Oops" in str(e), True)
        else:
            self.fail()

    def testEndVendorParsing(self):
        self.dict.ReadDictionary(
            StringIO(
                "VENDOR Simplon 42\n"
                "BEGIN-VENDOR Simplon\n"
                "END-VENDOR Simplon\n"
                "ATTRIBUTE Test-Type 1 integer"
            )
        )
        self.assertEqual(self.dict.attrindex["Test-Type"], 1)

    def testInclude(self):
        try:
            self.dict.ReadDictionary(
                StringIO(
                    "$INCLUDE this_file_does_not_exist\n"
                    "VENDOR Simplon 42\n"
                    "BEGIN-VENDOR Simplon\n"
                    "END-VENDOR Simplon\n"
                    "ATTRIBUTE Test-Type 1 integer"
                )
            )
        except OSError as e:
            self.assertEqual("this_file_does_not_exist" in str(e), True)
        else:
            self.fail()

    def testDictFilePostParse(self):
        f = DictFile(StringIO("VENDOR Simplon 42\n"))
        for _ in f:
            pass
        self.assertEqual(f.File(), "")
        self.assertEqual(f.Line(), -1)

    def testDictFileParseError(self):
        tmpdict = Dictionary()
        try:
            tmpdict.ReadDictionary(os.path.join(self.path, "dictfiletest"))
        except ParseError as e:
            self.assertEqual("dictfiletest" in str(e), True)
        else:
            self.fail()



import hashlib
import os
if os.name == 'nt':
    import selectors
else:
    import select
import socket
import struct
import time
from typing import Optional

from pyrad2 import host, packet
from pyrad2.constants import PacketType, EAPPacketType, EAPType
from pyrad2.dictionary import Dictionary
from pyrad2.exceptions import Timeout


class Client(host.Host):
    """Basic RADIUS client.
    This class implements a basic RADIUS client. It can send requests
    to a RADIUS server, taking care of timeouts and retries, and
    validate its replies.
    """

    def __init__(
        self,
        server: str,
        authport: int = 1812,
        acctport: int = 1813,
        coaport: int = 3799,
        secret: bytes = b"",
        dict: Optional[Dictionary] = None,
        retries: int = 3,
        timeout: int = 5,
    ):
        """Initializes a RADIUS client.

        Args:
            server (str): Hostname or IP address of the RADIUS server.
            authport (int): Port to use for authentication packets.
            acctport (int): Port to use for accounting packets.
            coaport (int): Port to use for CoA packets.
            secret (bytes): RADIUS secret.
            dict (pyrad.dictionary.Dictionary): RADIUS dictionary.
            retries (int): Number of times to retry sending a RADIUS request.
            timeout (int): Number of seconds to wait for an answer.
        """
        super().__init__(authport, acctport, coaport, dict)

        self.server = server
        self.secret = secret
        self.retries = retries
        self.timeout = timeout
        if os.name == 'nt':
            self._sel = selectors.DefaultSelector()
        else:
            self._poll = select.poll()
        self._socket: Optional[socket.socket] = None

    def bind(self, addr: str | tuple) -> None:
        """Bind socket to an address.
        Binding the socket used for communicating to an address can be
        usefull when working on a machine with multiple addresses.

        Args:
            addr (str | tuple): network address (hostname or IP) and port to bind to
        """
        self._CloseSocket()
        self._SocketOpen()
        if self._socket:
            self._socket.bind(addr)
        else:
            raise RuntimeError("No socket present")

    def _SocketOpen(self) -> None:
        try:
            family = socket.getaddrinfo(self.server, 80)[0][0]
        except Exception:
            family = socket.AF_INET
        if not self._socket:
            self._socket = socket.socket(family, socket.SOCK_DGRAM)
            self._socket.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
            if os.name == 'nt':
                self._sel.register(self._socket, selectors.EVENT_READ)
            else:
                self._poll.register(self._socket, select.POLLIN)

    def _CloseSocket(self) -> None:
        if self._socket:
            if os.name == 'nt':
                self._sel.unregister(self._socket)
            else:
                self._poll.unregister(self._socket)
            self._socket.close()
            self._socket = None

    def CreateAuthPacket(self, **args) -> packet.Packet:
        """Create a new RADIUS packet.
        This utility function creates a new RADIUS packet which can
        be used to communicate with the RADIUS server this client
        talks to. This is initializing the new packet with the
        dictionary and secret used for the client.

        Returns:
            packet.Packet: A new empty packet instance
        """
        return super().CreateAuthPacket(secret=self.secret, **args)

    def CreateAcctPacket(self, **args) -> packet.Packet:
        """Create a new RADIUS packet.
        This utility function creates a new RADIUS packet which can
        be used to communicate with the RADIUS server this client
        talks to. This is initializing the new packet with the
        dictionary and secret used for the client.

        Returns:
            packet.Packet: A new empty packet instance
        """
        return super().CreateAcctPacket(secret=self.secret, **args)

    def CreateCoAPacket(self, **args) -> packet.Packet:
        """Create a new RADIUS packet.
        This utility function creates a new RADIUS packet which can
        be used to communicate with the RADIUS server this client
        talks to. This is initializing the new packet with the
        dictionary and secret used for the client.

        Returns:
            packet.Packet: A new empty packet instance
        """
        return super().CreateCoAPacket(secret=self.secret, **args)

    def _SendPacket(self, pkt: packet.PacketImplementation, port: int):
        """Send a packet to a RADIUS server.

        Args:
            pkt (packet.Packet): The packet to send
            port (int): UDP port to send packet to

        Returns:
            packet.Packet: The reply packet received

        Raises:
            Timeout: RADIUS server does not reply
        """
        self._SocketOpen()

        for attempt in range(self.retries):
            if attempt and pkt.code == PacketType.AccountingRequest:
                if "Acct-Delay-Time" in pkt:
                    pkt["Acct-Delay-Time"] = pkt["Acct-Delay-Time"][0] + self.timeout
                else:
                    pkt["Acct-Delay-Time"] = self.timeout

            now = time.time()
            waitto = now + self.timeout

            if not self._socket:
                raise RuntimeError("No socket present")

            self._socket.sendto(pkt.RequestPacket(), (self.server, port))

            while now < waitto:
                rawreply = None

                if os.name == 'nt':
                    for key, mask in self._sel.select(timeout=(waitto - now)):
                        if mask & selectors.EVENT_READ:
                            rawreply = key.fileobj.recv(4096)

                else:
                    ready = self._poll.poll((waitto - now) * 1000)

                    if ready:
                        rawreply = self._socket.recv(4096)

                if not rawreply:
                    now = time.time()
                    continue

                try:
                    reply = pkt.CreateReply(packet=rawreply)
                    if pkt.VerifyReply(reply, rawreply):
                        return reply
                except packet.PacketError:
                    pass

                now = time.time()

        raise Timeout

    def SendPacket(self, pkt: packet.PacketImplementation) -> packet.Packet:  # type: ignore
        """Send a packet to a RADIUS server.

        Args:
            pkt (packet.Packet): Packet to send

        Returns:
            packet.Packet: The reply packet received

        Raises:
            Timeout: RADIUS server does not reply
        """
        if isinstance(pkt, packet.AuthPacket):
            if pkt.auth_type == "eap-md5":
                # Creating EAP-Identity
                password = pkt[2][0] if 2 in pkt else pkt[1][0]
                pkt[79] = [
                    struct.pack(
                        "!BBHB%ds" % len(password),
                        EAPPacketType.RESPONSE,
                        packet.CurrentID,
                        len(password) + 5,
                        EAPType.IDENTITY,
                        password,
                    )
                ]
            reply = self._SendPacket(pkt, self.authport)
            if (
                reply
                and reply.code == PacketType.AccessChallenge
                and pkt.auth_type == "eap-md5"
            ):
                # Got an Access-Challenge
                eap_code, eap_id, eap_size, eap_type, eap_md5 = struct.unpack(
                    "!BBHB%ds" % (len(reply[79][0]) - 5), reply[79][0]
                )
                # Sending back an EAP-Type-MD5-Challenge
                # Thank god for http://www.secdev.org/python/eapy.py
                client_pw = pkt[2][0] if 2 in pkt else pkt[1][0]
                md5_challenge = hashlib.md5(
                    struct.pack("!B", eap_id) + client_pw + eap_md5[1:]
                ).digest()
                pkt[79] = [
                    struct.pack(
                        "!BBHBB",
                        2,
                        eap_id,
                        len(md5_challenge) + 6,
                        4,
                        len(md5_challenge),
                    )
                    + md5_challenge
                ]
                # Copy over Challenge-State
                pkt[24] = reply[24]
                reply = self._SendPacket(pkt, self.authport)
            return reply
        elif isinstance(pkt, packet.CoAPacket):
            return self._SendPacket(pkt, self.coaport)
        else:
            return self._SendPacket(pkt, self.acctport)



import asyncio
import ssl
from abc import abstractmethod
from typing import Optional

from loguru import logger

from pyrad2.constants import PacketType
from pyrad2.dictionary import Dictionary
from pyrad2.packet import (
    AcctPacket,
    AuthPacket,
    CoAPacket,
    Packet,
    PacketError,
    parse_packet,
)
from pyrad2.server import RemoteHost, ServerPacketError
from pyrad2.tools import (
    get_cert_fingerprint,
    read_radius_packet,
)


class UnknownHost(Exception):
    pass


class RadSecServer:
    """A RadSec as per RFC6614.

    UDP + MD5 has proven to be a combination that has not survived
    the test of time. Hence, the RADIUS standard adopted RADSEC
    as a fundamentally more secure approach.

    RADSEC effectively means performing communications over TCP instead of UDP
    (generally on port 2083) and use TLS as a security layer.

    RADSEC is the same as “Radius Over TLS” or Radius/TLS.

    The default destination port number for RADIUS over TLS is TCP/2083.
    There are no separate ports for authentication, accounting, and
    dynamic authorization changes.
    """

    ALLOWED_CIPHERS = "DES-CBC3-SHA:RC4-SHA:AES128-SHA"

    def __init__(
        self,
        listen_address: str = "0.0.0.0",
        listen_port: int = 2083,
        hosts: Optional[dict[str, RemoteHost]] = None,
        dictionary: Optional[Dictionary] = None,
        verify_packet: bool = False,
        certfile: str = "certs/server/server.cert.pem",
        keyfile: str = "certs/server/server.key.pem",
        ca_certfile: str = "certs/ca/ca.cert.pem",
        verify_mode: ssl.VerifyMode = ssl.CERT_NONE,
    ):
        """Initializes a RadSec server.

        Args:
            listen_address (str): IP address to bind to, defaults to 0.0.0.0
            listen_port (int): Deafaults to 2083.
            hosts (dict[str, RemoteHost]): Hosts who we can talk to. A dictionary mapping IP to RemoteHost class instances.
            dictionary (Dictionary): RADIUS dictionary to use.
            verify_packet (bool): If true, the packet will be verified against its secret
            certfile (str): Path to server SSL certificate
            keyfile (str): Path to server SSL certificate
            ca_certfile (str): Path to server CA certfificate
        """
        self.listen_address = listen_address
        self.listen_port = listen_port
        self.hosts = {} if hosts is None else hosts
        self.dict = dictionary
        self.verify_packet = verify_packet

        self.setup_ssl(certfile, keyfile, ca_certfile, verify_mode)

    async def run(self):
        server = await asyncio.start_server(
            self._handle_client,
            host=self.listen_address,
            port=self.listen_port,
            ssl=self.ssl_ctx,
        )

        addr = server.sockets[0].getsockname()
        logger.info("RADSEC Server with mutual TLS running on {}", addr)
        logger.info("Allowed ciphers: {}", self.ALLOWED_CIPHERS)

        try:
            async with server:
                await server.serve_forever()
        except asyncio.CancelledError:
            logger.info("Task cancelled")
        except KeyboardInterrupt:
            logger.info("Server killed manually")
        finally:
            server.close()
            await server.wait_closed()
            logger.info("Server shutdown")

    def setup_ssl(self, certfile: str, keyfile: str, ca_certfile: str, verify_mode):
        ssl_ctx = ssl.create_default_context(ssl.Purpose.CLIENT_AUTH)
        try:
            ssl_ctx.load_cert_chain(certfile=certfile, keyfile=keyfile)
        except FileNotFoundError as e:
            ssl_paths = ", ".join([certfile, keyfile, ca_certfile])
            msg = "One or more SSL files could not be found. Current paths: {}"
            logger.error(msg, ssl_paths)
            raise FileNotFoundError(msg.format(ssl_paths)) from e

        ssl_ctx.verify_mode = verify_mode
        ssl_ctx.load_verify_locations(cafile=ca_certfile)
        ssl_ctx.set_ciphers(self.ALLOWED_CIPHERS)

        self.ssl_ctx = ssl_ctx

    async def _handle_client(
        self, reader: asyncio.StreamReader, writer: asyncio.StreamWriter
    ) -> None:
        peername = writer.get_extra_info("peername")
        cert_bin = writer.get_extra_info("peercert", default=None)

        client_id = None
        if cert_bin:
            client_id = writer.get_extra_info("ssl_object").getpeercert(
                binary_form=True
            )
            logger.info(
                "Client {} fingerprint: {}", peername, get_cert_fingerprint(client_id)
            )
        else:
            logger.warning("No certificate from client {}", peername)

        logger.info("RADSEC connection established from {}", peername)

        data = await read_radius_packet(reader)
        logger.info("Received {} bytes from {}", len(data), peername)
        logger.debug("Data (hex): {}", data.hex())

        try:
            reply = await self.packet_received(data, host=peername[0])
        except UnknownHost:
            logger.warning("Drop package from unknown source {}", peername[0])
            return

        writer.write(reply.ReplyPacket())
        await writer.drain()
        logger.info("Sent reply to {}: {}", peername, reply.code)

    async def packet_received(self, data: bytes, host: str) -> Packet:
        if host in self.hosts:
            remote_host = self.hosts[host]
        elif "0.0.0.0" in self.hosts:
            remote_host = self.hosts["0.0.0.0"]
        else:
            raise UnknownHost

        packet = parse_packet(data, remote_host.secret, self.dict)

        if self.verify_packet:
            if not packet.verify():
                raise PacketError("Packet verification failed")

        if packet.code == PacketType.AccessRequest:
            return await self.handle_access_request(packet)
        elif packet.code in (
            PacketType.AccountingRequest,
            PacketType.AccountingResponse,
        ):
            return await self.handle_accounting(packet)
        elif packet.code == PacketType.CoARequest:
            return await self.handle_coa(packet)
        elif packet.code == PacketType.DisconnectRequest:
            return await self.handle_disconnect(packet)
        else:
            raise ServerPacketError("Unsupported packet code: {}".format(packet.code))

    @abstractmethod
    async def handle_access_request(self, packet: AuthPacket) -> Packet:
        """Handle an Access-Request packet."""
        raise NotImplementedError("Subclasses must implement this method")

    @abstractmethod
    async def handle_accounting(self, packet: AcctPacket) -> Packet:
        """Handle an Accounting-Request or Accounting-Response packet."""
        raise NotImplementedError("Subclasses must implement this method")

    @abstractmethod
    async def handle_coa(self, packet: CoAPacket) -> Packet:
        """Handle a CoA-Request packet."""
        raise NotImplementedError("Subclasses must implement this method")

    @abstractmethod
    async def handle_disconnect(self, packet: CoAPacket) -> Packet:
        """Handle a Disconnect-Request packet."""
        raise NotImplementedError("Subclasses must implement this method")



class RadiusException(Exception):
    pass


class ServerPacketError(RadiusException):
    """Exception class for bogus packets.
    ServerPacketError exceptions are only used inside the Server class to
    abort processing of a packet.
    """

    pass


class Timeout(RadiusException):
    """Simple exception class which is raised when a timeout occurs
    while waiting for a RADIUS server to respond."""

    pass


class PacketError(RadiusException):
    """Raised when the packet is invalid."""

    pass


class ParseError(RadiusException):
    """Exception raised for errors
    while parsing RADIUS dictionary files.

    Attributes:
        msg (str): Error message
        linenumber (int): Line number on which the error occurred
    """

    def __init__(self, msg=None, **data):
        self.msg = msg
        self.file = data.get("file", "")
        self.line = data.get("line", -1)

    def __str__(self):
        str = ""
        if self.file:
            str += self.file
        if self.line > -1:
            str += "(%d)" % self.line
        if self.file or self.line > -1:
            str += ": "
        str += "Parse error"
        if self.msg:
            str += ": %s" % self.msg

        return str
